from Components.Label import Label
from Components.config import config, ConfigInteger, ConfigSubsection
from Screens.Screen import Screen
from Screens.MessageBox import MessageBox
from Components.ActionMap import ActionMap
from Plugins.Plugin import PluginDescriptor
from Components.ScrollLabel import ScrollLabel  # For scrollable log text
from enigma import eTimer
import psutil
import os

# ✅ Config for CPU threshold
config.plugins.CPUUsageMonitor = ConfigSubsection()
config.plugins.CPUUsageMonitor.threshold = ConfigInteger(default=50, limits=(0, 100))

cpu_checker_instance = None  # Global instance to control logging

class CPUUsageMonitor(Screen):
    skin = """
        <screen name="CPUUsageMonitor" position="center,center" size="1300,850" title="CPU + Memory Monitor">
            <widget name="info" position="20,20" size="1260,60" font="Regular;40" foregroundColor="#00ffc000"/>
            <widget name="status" position="20,100" size="1260,60" font="Regular;34" transparent="0" foregroundColor="#FFFFFF" />
            <widget name="threshold_display" position="20,180" size="1260,60" font="Regular;34" foregroundColor="#FFD700"/>
            <widget name="cpu_usage" position="20,260" size="1260,60" font="Regular;34" foregroundColor="#00FFFF" text="CPU Usage: -"/>
            <widget name="memory_usage" position="20,320" size="1260,60" font="Regular;34" foregroundColor="#FF69B4" text="Memory Usage: -"/>

            <ePixmap pixmap="skin_default/buttons/green.png" position="50,650" size="250,70" alphatest="on" />
            <widget name="key_green" position="50,730" size="250,50" valign="center" halign="center" font="Regular;32" foregroundColor="#00FF00" transparent="0" text="Start" />

            <ePixmap pixmap="skin_default/buttons/red.png" position="350,650" size="250,70" alphatest="on" />
            <widget name="key_red" position="350,730" size="250,50" valign="center" halign="center" font="Regular;32" foregroundColor="#FF0000" transparent="0" text="Stop" />

            <ePixmap pixmap="skin_default/buttons/yellow.png" position="650,650" size="250,70" alphatest="on" />
            <widget name="key_yellow" position="650,730" size="250,50" valign="center" halign="center" font="Regular;32" foregroundColor="#FFFF00" transparent="0" text="Clear Log" />

            <ePixmap pixmap="skin_default/buttons/blue.png" position="950,650" size="250,70" alphatest="on" />
            <widget name="key_blue" position="950,730" size="250,50" valign="center" halign="center" font="Regular;32" foregroundColor="#0000FF" transparent="0" text="Exit" />
        </screen>
    """

    def __init__(self, session):
        Screen.__init__(self, session)
        self.session = session
        self.threshold = config.plugins.CPUUsageMonitor.threshold.value  # Store threshold locally

        # Labels for UI
        self["info"] = Label("CPU & Memory Monitor")
        self["status"] = Label("Status: Checking...")
        self["threshold_display"] = Label(f"Threshold: {self.threshold}%")
        self["cpu_usage"] = Label("CPU Usage: -")
        self["memory_usage"] = Label("Memory Usage: -")
        
        self.timer = eTimer()
        self.timer.callback.append(self.updateUsage)
        self.timer.start(5000, True)  # Refresh every 5 seconds

        self.updateStatus()  # Check and set correct status on launch

        self["key_green"] = Label("Start")
        self["key_red"] = Label("Stop")
        self["key_yellow"] = Label("Clear Log")
        self["key_blue"] = Label("Exit")

        # Key Bindings (Color & Setup Actions)
        self["actions"] = ActionMap(["ColorActions", "SetupActions", "MenuActions"], {
            "green": self.startLogging,
            "red": self.stopLogging,
            "yellow": self.clearLog,
            "blue": self.close,
            "cancel": self.close,  # Allow "EXIT" button to work
            "left": self.decreaseThreshold,
            "right": self.increaseThreshold,
            "menu": self.openLogViewer  # New key mapping to open the log viewer
        }, -2)
    
    def updateUsage(self):
        """Update CPU & Memory usage on screen and show bold if above threshold"""
        cpu_usage = psutil.cpu_percent()
        mem_usage = psutil.virtual_memory().percent

        # CPU usage display (bold if above threshold)
        if cpu_usage > self.threshold:
            self["cpu_usage"].setText(f"CPU Usage: {cpu_usage}%")
            try:
                self["cpu_usage"].instance.setFont("Regular Bold;34")
            except Exception:
                pass
        else:
            self["cpu_usage"].setText(f"CPU Usage: {cpu_usage}%")
            try:
                self["cpu_usage"].instance.setFont("Regular;34")
            except Exception:
                pass

        # Memory usage display (bold if above threshold)
        if mem_usage > self.threshold:
            self["memory_usage"].setText(f"Memory Usage: {mem_usage}%")
            try:
                self["memory_usage"].instance.setFont("Regular Bold;34")
            except Exception:
                pass
        else:
            self["memory_usage"].setText(f"Memory Usage: {mem_usage}%")
            try:
                self["memory_usage"].instance.setFont("Regular;34")
            except Exception:
                pass

        self.timer.start(5000, True)  # Refresh every 5 seconds
    
    def updateStatus(self):
        """Check if logging is running and update status"""
        global cpu_checker_instance
        if cpu_checker_instance and cpu_checker_instance.logging:
            self["status"].setText("Status: Logging Running")
        else:
            self["status"].setText("Status: Not Running")

    def increaseThreshold(self):
        """Increase CPU Threshold & Update Display"""
        if self.threshold < 100:
            self.threshold += 5
            config.plugins.CPUUsageMonitor.threshold.value = self.threshold
            config.plugins.CPUUsageMonitor.threshold.save()
            self["threshold_display"].setText(f"Threshold: {self.threshold}%")
        else:
            self.session.open(MessageBox, "Max threshold reached!", MessageBox.TYPE_INFO)

    def decreaseThreshold(self):
        """Decrease CPU Threshold & Update Display"""
        if self.threshold > 0:
            self.threshold -= 5
            config.plugins.CPUUsageMonitor.threshold.value = self.threshold
            config.plugins.CPUUsageMonitor.threshold.save()
            self["threshold_display"].setText(f"Threshold: {self.threshold}%")
        else:
            self.session.open(MessageBox, "Min threshold reached!", MessageBox.TYPE_INFO)

    def startLogging(self):
        global cpu_checker_instance
        if cpu_checker_instance is None:
            cpu_checker_instance = CPUUsageChecker(self.session)
            cpu_checker_instance.startLogging()
            self["status"].setText("Status: Logging Started")
        else:
            if not cpu_checker_instance.logging:
                cpu_checker_instance.session = self.session
                cpu_checker_instance.startLogging()
                self["status"].setText("Status: Logging Restarted")
            else:
                self.session.open(MessageBox, "Logging already running!", MessageBox.TYPE_INFO)

    def stopLogging(self):
        global cpu_checker_instance
        if cpu_checker_instance:
            cpu_checker_instance.stopLogging()
            self["status"].setText("Status: Logging Stopped")
        else:
            self.session.open(MessageBox, "No active logging session!", MessageBox.TYPE_INFO)

    def clearLog(self):
        """Clear log file WITHOUT deleting it"""
        log_file = "/tmp/cpu_monitor.log"
    
        try:
            with open(log_file, 'w') as log:
                log.write("🗑 Log Cleared\n")
            self.session.open(MessageBox, "Log file cleared!", MessageBox.TYPE_INFO)
            print("🗑 Log Cleared Successfully")
        except Exception as e:
            self.session.open(MessageBox, f"Error clearing log: {e}", MessageBox.TYPE_WARNING)
            print(f"❌ Error clearing log: {e}")

    def openLogViewer(self):
        """Open a new screen to view the log file with scrolling support"""
        self.session.open(LogViewer)

    def close(self):
        """Close plugin window but KEEP logging running"""
        print("🔄 Closing Plugin Window (Logging Continues)")
        Screen.close(self)

class LogViewer(Screen):
    skin = """
        <screen name="LogViewer" position="center,center" size="1300,950" title="CPU Monitor Log">
            <widget name="logText" position="20,20" size="1260,760" font="Regular;30" scrollbarMode="showAlways" />

            <!-- Blue Button (Exit) -->
            <ePixmap pixmap="skin_default/buttons/blue.png" position="50,800" size="150,50" alphatest="on" />
            <widget name="key_blue" position="50,800" size="150,50" valign="center" halign="center"
                font="Regular;32" foregroundColor="#FFFFFF" transparent="0" text="Exit" />

            <!-- Yellow Button (Clear Log) -->
            <ePixmap pixmap="skin_default/buttons/yellow.png" position="220,800" size="150,50" alphatest="on" />
            <widget name="key_yellow" position="220,800" size="150,50" valign="center" halign="center"
                font="Regular;32" foregroundColor="#FFFF00" transparent="0" text="Clear Log" />
        </screen>
    """

    LOG_FILE = "/tmp/cpu_monitor.log"

    def __init__(self, session):
        Screen.__init__(self, session)
        self.session = session

        self["logText"] = ScrollLabel("Loading log...")
        self["key_blue"] = Label("Exit")
        self["key_yellow"] = Label("Clear Log")

        self["actions"] = ActionMap(["WizardActions", "ColorActions", "DirectionActions"], {
            "blue": self.close,
            "yellow": self.clearLog,
            "cancel": self.close,  # Allow EXIT button to work
            "ok": self.close,      # Allow OK button to work
            "up": self.pageUp,
            "down": self.pageDown
        }, -2)

        self.logTimer = eTimer()
        self.logTimer.callback.append(self.updateLog)

        self.updateLog(initial=True)
        self.logTimer.start(3000, True)

    def updateLog(self, initial=False):
        """ Continuously read new lines from the log file and update the display """
        if not os.path.exists(self.LOG_FILE):
            self["logText"].setText("⚠ Log file not found!")
            return

        try:
            with open(self.LOG_FILE, "r") as f:
                log_content = f.read()
                if log_content:
                    self["logText"].setText(log_content)
                    self["logText"].lastPage()

        except Exception as e:
            self["logText"].setText(f"⚠ Error reading log file: {str(e)}")

        self.logTimer.start(3000, True)

    def clearLog(self):
        """ Clear the log file and reset the display """
        try:
            with open(self.LOG_FILE, 'w') as log:
                log.write("🗑 Log Cleared\n")

            self.session.open(MessageBox, "Log file cleared!", MessageBox.TYPE_INFO)
            self["logText"].setText("🗑 Log Cleared\n")

        except Exception as e:
            self.session.open(MessageBox, f"Error clearing log: {e}", MessageBox.TYPE_WARNING)

    def pageUp(self):
        """ Scroll up in log viewer """
        if self["logText"]:
            self["logText"].pageUp()

    def pageDown(self):
        """ Scroll down in log viewer """
        if self["logText"]:
            self["logText"].pageDown()

    def close(self):
        """ Stop the auto-refresh timer and close the log viewer """
        if hasattr(self, "logTimer"):
            self.logTimer.stop()
        Screen.close(self)



class CPUUsageChecker:
    instance = None  # Singleton instance
    logging = False

    @classmethod
    def getInstance(cls, session=None):
        """Ensure only ONE instance of CPUUsageChecker exists"""
        if cls.instance is None:
            cls.instance = cls(session)
        return cls.instance

    def __init__(self, session):
        if CPUUsageChecker.instance:
            return  # Prevent duplicate instances
        CPUUsageChecker.instance = self
        self.session = session  # Store session correctly
        self.timer = eTimer()
        self.timer.callback.append(self.logUsage)
        print("✅ CPUUsageChecker initialized!")

    def startLogging(self):
        """Start CPU/Memory logging"""
        if not self.logging:
            self.logging = True

            # Ensure timer is initialized before starting
            if not hasattr(self, "timer"):
                self.timer = eTimer()
                self.timer.callback.append(self.logUsage)

            if not hasattr(self, "session") or self.session is None:
                self.session = CPUUsageMonitor.instance.session

            with open("/tmp/cpu_monitor.log", "a") as log:
                log.write("✅ Logging Started\n")

            self.timer.start(10000, False)  # Log every 10 seconds
            print("✅ CPU/Memory Logging Started")
            self.logUsage()  # First log entry immediately

    def stopLogging(self):
        """Stop CPU/Memory logging"""
        if self.logging:
            self.logging = False
            if hasattr(self, "timer"):
                self.timer.stop()

            with open("/tmp/cpu_monitor.log", "a") as log:
                log.write("🛑 Logging Stopped\n")

            print("🛑 CPU/Memory Logging Stopped")

    def stop(self):
        """Stop CPU/Memory logging (additional stop method)"""
        if self.logging:
            self.logging = False
            self.timer.stop()
            print("🛑 CPU/Memory Logging Stopped")

    def logUsage(self):
        """Log CPU & Memory, check thresholds, and restart timer"""
        if not self.logging:
            return

        cpu_usage = psutil.cpu_percent()
        mem_usage = psutil.virtual_memory().percent
        threshold = config.plugins.CPUUsageMonitor.threshold.value

        with open("/tmp/cpu_monitor.log", "a") as log:
            log.write(f"CPU: {cpu_usage}% | Memory: {mem_usage}% (Threshold: {threshold}%)\n")

        if self.session and (cpu_usage > threshold or mem_usage > threshold):
            self.session.open(MessageBox, f"⚠ High Usage! CPU: {cpu_usage}% | Memory: {mem_usage}%", MessageBox.TYPE_WARNING)

        self.timer.start(10000, False)  # Restart timer automatically

def main(session, **kwargs):
    session.open(CPUUsageMonitor)

def Plugins(**kwargs):
    return [
        PluginDescriptor(name="CPU Usage Monitor", description="Monitor CPU & Memory Usage", where=PluginDescriptor.WHERE_PLUGINMENU, fnc=main)
    ]
